#ifndef BqW_autoptr_h
#define BqW_autoptr_h

/* This is Bisqwit's generic autoptr.hh.
 * The same file is used in many different projects.
 *
 * autoptr.hh version 1.1.0
 */

template <typename T>
class autoptr
{
    T *p;
    void Del() { if(p && !--p->refnum)delete p; }
    void Set(T *a) { if(a)++a->refnum; if(p && !--p->refnum)delete p; p = a; }
public:
    autoptr() : p(NULL) { }
    autoptr(T *const a) { p=a; if(p)++p->refnum; }
    autoptr &operator= (T *const a) { Set(a); return *this; }
    autoptr &operator= (const autoptr<T> &a) { Set(&*a); return *this; }
    T &operator* () const { return *p; }
    T *operator-> () const { return p; }
    // to enable boolean tests with this class: if(tmp) {...}
    operator bool() const { return p; }
    operator T *() const { return p; }
    operator const T *() const { return p; }
    bool operator! () const { return !p; }
    ~autoptr() { Del(); }
    autoptr(const autoptr<T> &a) { p=&*a; if(p)++p->refnum; }
};

template <typename T>
class const_autoptr
{
    const T *p;
    void Del() { if(p && !--p->refnum)delete p; }
    void Set(const T *a) { if(a)++a->refnum; if(p && !--p->refnum)delete p; p = a; }
public:
    const_autoptr() : p(NULL) { }
    const_autoptr(const T *const a) { p=a; if(p)++p->refnum; }
    const_autoptr &operator= (const T *const a) { Set(a); return *this; }
    const_autoptr &operator= (const const_autoptr<T> &a) { Set(&*a); return *this; }
    const_autoptr &operator= (const autoptr<T> &a) { Set(&*a); return *this; }
    const T &operator* () const { return *p; }
    const T *operator-> () const { return p; }
    // to enable boolean tests with this class: if(tmp) {...}
    operator const T *() { return p; }
    bool operator! () const { return !p; }
    ~const_autoptr() { Del(); }
    const_autoptr(const const_autoptr<T> &a) { p=&*a; if(p)++p->refnum; }
    const_autoptr(const autoptr<T> &a) { p=&*a; if(p)++p->refnum; }
};

// autoptr must only use ptrable classes.
class ptrable
{
    // This does not compile with some old
    // egcs and gcc compilers. Don't blame
    // me, update your buggy compiler.
    template<typename T> friend autoptr<T>;
    template<typename T> friend const_autoptr<T>;
    mutable int refnum;
public:
    ptrable() : refnum(0) { }
};

#endif
